class ServiceException(Exception):
    """
    Base class for Service exceptions
    """

    pass


class LifecycleError(ServiceException):
    """
    Raised when an action would violate the service lifecycle rules.
    """

    pass


class DaemonTaskExit(ServiceException):
    """
    Raised when an action would violate the service lifecycle rules.
    """

    pass


class ServiceCancelled(ServiceException):
    """
    Raised by external service APIs when a service is cancelled.
    """

    pass



import itertools
from typing import Collection, Iterable, Mapping, Set, TypeVar

TItem = TypeVar("TItem")


def iter_dag(dag: Mapping[TItem, Collection[TItem]]) -> Iterable[TItem]:
    """
    Iterate over a DAG.

    The DAG is structured as a mapping of `parent -> children` such that each
    children is a set of elements that depend on `parent`.

    The iteration happens such that for any `parent` that is being yielded, all
    of the `children` for that parent will have been previosly yielded.
    """
    seen: Set[TItem] = set()
    all_items = set(dag.keys())
    all_children = set(itertools.chain.from_iterable(dag.values()))

    if not all_items.issuperset(all_children):
        # Ensure that
        raise ValueError("Parents not representative of all children")

    while seen != all_items:
        num_seen = len(seen)
        for item in dag.keys():
            if item in seen:
                # already yielded item
                continue
            elif seen.issuperset(dag[item]):
                # all children have been yielded so yield item
                yield item
                seen.add(item)
        if len(seen) == num_seen:
            raise Exception(
                "Unable to iterate over DAG. This can happen if the data "
                "structure is not actaully a dag and contains a cyclic "
                "dependency."
            )



from abc import ABC, abstractmethod
from typing import Any, Awaitable, Callable

import trio_typing


class ServiceAPI(ABC):
    manager: "_InternalManagerAPI"

    @abstractmethod
    async def run(self) -> None:
        """
        Primary entry point for all service logic.

        .. note:: This method should **not** be directly invoked by user code.

        Services may be run using the following approaches.

        .. code-block: python

            # 1. run the service in the background using a context manager
            async with run_service(service) as manager:
                # service runs inside context block
                ...
                # service cancels and stops when context exits
            # service will have fully stopped

            # 2. run the service blocking until completion
            await Manager.run_service(service)

            # 3. create manager and then run service blocking until completion
            manager = Manager(service)
            await manager.run()
        """
        ...


class ManagerAPI(ABC):
    @property
    @abstractmethod
    def is_started(self) -> bool:
        """
        Return boolean indicating if the underlying service has been started.
        """
        ...

    @property
    @abstractmethod
    def is_running(self) -> bool:
        """
        Return boolean indicating if the underlying service is actively
        running.

        A service is considered running if it has been started and
        has not yet been stopped.
        """
        ...

    @property
    @abstractmethod
    def is_cancelled(self) -> bool:
        """
        Return boolean indicating if the underlying service has been cancelled.

        This can occure externally via the `cancel()` method or internally due
        to a task crash or a crash of the actual :meth:`ServiceAPI.run` method.
        """
        ...

    @property
    @abstractmethod
    def is_stopping(self) -> bool:
        """
        Return boolean indicating if the underlying service is in the process
        of shutting down.
        """
        ...

    @property
    @abstractmethod
    def is_finished(self) -> bool:
        """
        Return boolean indicating if the underlying service is stopped.

        A stopped service will have completed all of the background tasks.
        """
        ...

    @property
    @abstractmethod
    def did_error(self) -> bool:
        """
        Return boolean indicating if the underlying service threw an exception.
        """
        ...

    @abstractmethod
    def cancel(self) -> None:
        """
        Trigger cancellation of the service.
        """
        ...

    @abstractmethod
    async def stop(self) -> None:
        """
        Trigger cancellation of the service and wait for it to finish.
        """
        ...

    @abstractmethod
    async def wait_started(self) -> None:
        """
        Wait until the service is started.
        """
        ...

    @abstractmethod
    async def wait_stopping(self) -> None:
        """
        Wait until the service is shutting down.
        """
        ...

    @abstractmethod
    async def wait_finished(self) -> None:
        """
        Wait until the service is stopped.
        """
        ...

    @classmethod
    @abstractmethod
    async def run_service(cls, service: ServiceAPI) -> None:
        """
        Run a service
        """
        ...

    @abstractmethod
    async def run(self) -> None:
        """
        Run a service
        """
        ...


class _InternalManagerAPI(ManagerAPI):
    """
    Defines the API that the `Service.manager` property exposes.

    The _InternalManagerAPI / ManagerAPI distinction is in place to ensure that
    external callers to a service do not try to use the task scheduling
    functionality as it is only designed to be used internally.
    """

    @trio_typing.takes_callable_and_args
    @abstractmethod
    def run_task(
        self,
        async_fn: Callable[..., Awaitable[Any]],
        *args: Any,
        daemon: bool = False,
        name: str = None
    ) -> None:
        """
        Run a task in the background.  If the function throws an exception it
        will trigger the service to be cancelled and be propogated.

        If `daemon == True` then the the task is expected to run indefinitely
        and will trigger cancellation if the task finishes.
        """
        ...

    @trio_typing.takes_callable_and_args
    @abstractmethod
    def run_daemon_task(
        self, async_fn: Callable[..., Awaitable[Any]], *args: Any, name: str = None
    ) -> None:
        """
        Run a daemon task in the background.

        Equivalent to `run_task(..., daemon=True)`.
        """
        ...

    @abstractmethod
    def run_child_service(
        self, service: ServiceAPI, daemon: bool = False, name: str = None
    ) -> "ManagerAPI":
        """
        Run a service in the background.  If the function throws an exception it
        will trigger the parent service to be cancelled and be propogated.

        If `daemon == True` then the the service is expected to run indefinitely
        and will trigger cancellation if the service finishes.
        """
        ...

    @abstractmethod
    def run_daemon_child_service(
        self, service: ServiceAPI, name: str = None
    ) -> "ManagerAPI":
        """
        Run a daemon service in the background.

        Equivalent to `run_child_service(..., daemon=True)`.
        """
        ...



import functools
import sys
from typing import Any, AsyncIterator, Awaitable, Callable, cast

from async_generator import asynccontextmanager
import trio
import trio_typing

from .abc import ManagerAPI, ServiceAPI
from .base import BaseManager
from .exceptions import DaemonTaskExit, LifecycleError
from .typing import EXC_INFO


class TrioManager(BaseManager):
    # A nursery for system tasks.  This nursery is cancelled in the event that
    # the service is cancelled or exits.
    _system_nursery: trio_typing.Nursery

    # A nursery for sub tasks and services.  This nursery is cancelled if the
    # service is cancelled but allowed to exit normally if the service exits.
    _task_nursery: trio_typing.Nursery

    def __init__(self, service: ServiceAPI) -> None:
        super().__init__(service)

        # events
        self._started = trio.Event()
        self._cancelled = trio.Event()
        self._stopping = trio.Event()
        self._finished = trio.Event()

        # locks
        self._run_lock = trio.Lock()

    #
    # System Tasks
    #
    async def _handle_cancelled(self, task_nursery: trio_typing.Nursery) -> None:
        """
        Handle cancellation of the task nursery.
        """
        self.logger.debug("%s: _handle_cancelled waiting for cancellation", self)
        await self.wait_cancelled()
        self.logger.debug(
            "%s: _handle_cancelled triggering task nursery cancellation", self
        )
        task_nursery.cancel_scope.cancel()

    async def _handle_stopping(self, system_nursery: trio_typing.Nursery) -> None:
        """
        Handle cancellation of the system nursery.
        """
        self.logger.debug("%s: _handle_stopping waiting for stopping", self)
        await self.wait_stopping()
        self.logger.debug(
            "%s: _handle_stopping triggering system nursery cancellation", self
        )
        system_nursery.cancel_scope.cancel()

    async def _handle_run(self) -> None:
        """
        Run and monitor the actual :meth:`ServiceAPI.run` method.

        In the event that it throws an exception the service will be cancelled.

        Upon a clean exit
        Triggers cancellation in the case where the service exits normally or
        throws an exception.
        """
        try:
            await self._service.run()
        except Exception as err:
            self.logger.debug(
                "%s: _handle_run got error, storing exception and setting cancelled: %s",
                self,
                err,
            )
            self._errors.append(cast(EXC_INFO, sys.exc_info()))
            self.cancel()
        else:
            # NOTE: Any service which uses daemon tasks will need to trigger
            # cancellation in order for the service to exit since this code
            # path does not trigger task cancellation.  It might make sense to
            # trigger cancellation if all of the running tasks are daemon
            # tasks.
            self.logger.debug(
                "%s: _handle_run exited cleanly, waiting for full stop...", self
            )

    @classmethod
    async def run_service(cls, service: ServiceAPI) -> None:
        manager = cls(service)
        await manager.run()

    async def run(self) -> None:
        if self._run_lock.locked():
            raise LifecycleError(
                "Cannot run a service with the run lock already engaged.  Already started?"
            )
        elif self.is_started:
            raise LifecycleError("Cannot run a service which is already started.")

        try:
            async with self._run_lock:
                async with trio.open_nursery() as system_nursery:
                    try:
                        async with trio.open_nursery() as task_nursery:
                            self._task_nursery = task_nursery

                            system_nursery.start_soon(
                                self._handle_cancelled, task_nursery
                            )
                            system_nursery.start_soon(
                                self._handle_stopping, system_nursery
                            )

                            task_nursery.start_soon(self._handle_run)

                            self._started.set()

                            # ***BLOCKING HERE***
                            # The code flow will block here until the background tasks have
                            # completed or cancellation occurs.
                    finally:
                        # signal that the service is stopping
                        self._stopping.set()
                        self.logger.debug("%s stopping", self)
        finally:
            self._finished.set()
            self.logger.debug("%s finished", self)

        # If an error occured, re-raise it here
        if self.did_error:
            raise trio.MultiError(
                tuple(
                    exc_value.with_traceback(exc_tb)
                    for _, exc_value, exc_tb in self._errors
                )
            )

    #
    # Event API mirror
    #
    @property
    def is_started(self) -> bool:
        return self._started.is_set()

    @property
    def is_cancelled(self) -> bool:
        return self._cancelled.is_set()

    @property
    def is_stopping(self) -> bool:
        return self._stopping.is_set() and not self.is_finished

    @property
    def is_finished(self) -> bool:
        return self._finished.is_set()

    #
    # Control API
    #
    def cancel(self) -> None:
        if not self.is_started:
            raise LifecycleError("Cannot cancel as service which was never started.")
        self._cancelled.set()

    #
    # Wait API
    #
    async def wait_started(self) -> None:
        await self._started.wait()

    async def wait_cancelled(self) -> None:
        await self._cancelled.wait()

    async def wait_stopping(self) -> None:
        await self._stopping.wait()

    async def wait_finished(self) -> None:
        await self._finished.wait()

    async def _run_and_manage_task(
        self,
        async_fn: Callable[..., Awaitable[Any]],
        *args: Any,
        daemon: bool,
        name: str,
    ) -> None:
        try:
            await async_fn(*args)
        except Exception as err:
            self.logger.debug(
                "task '%s[daemon=%s]' exited with error: %s",
                name,
                daemon,
                err,
                exc_info=True,
            )
            self._errors.append(cast(EXC_INFO, sys.exc_info()))
            self.cancel()
        else:
            self.logger.debug("task '%s[daemon=%s]' finished.", name, daemon)
            if daemon:
                self.logger.debug(
                    "daemon task '%s' exited unexpectedly.  Cancelling service: %s",
                    name,
                    self,
                )
                self.cancel()
                raise DaemonTaskExit(f"Daemon task {name} exited")

    def run_task(
        self,
        async_fn: Callable[..., Awaitable[Any]],
        *args: Any,
        daemon: bool = False,
        name: str = None,
    ) -> None:

        self._task_nursery.start_soon(
            functools.partial(
                self._run_and_manage_task, daemon=daemon, name=name or repr(async_fn)
            ),
            async_fn,
            *args,
            name=name,
        )

    def run_child_service(
        self, service: ServiceAPI, daemon: bool = False, name: str = None
    ) -> ManagerAPI:
        child_manager = type(self)(service)
        self.run_task(child_manager.run, daemon=daemon, name=name or repr(service))
        return child_manager


@asynccontextmanager
async def background_trio_service(service: ServiceAPI) -> AsyncIterator[ManagerAPI]:
    """
    Run a service in the background.

    The service is running within the context
    block and will be properly cleaned up upon exiting the context block.
    """
    async with trio.open_nursery() as nursery:
        manager = TrioManager(service)
        nursery.start_soon(manager.run)
        await manager.wait_started()
        try:
            yield manager
        finally:
            await manager.stop()



from .abc import ManagerAPI, ServiceAPI  # noqa: F401
from .asyncio import AsyncioManager, background_asyncio_service  # noqa: F401
from .base import Service, as_service  # noqa: F401
from .exceptions import DaemonTaskExit, LifecycleError, ServiceCancelled  # noqa: F401
from .trio import TrioManager, background_trio_service  # noqa: F401

run_asyncio_service = AsyncioManager.run_service
run_trio_service = TrioManager.run_service



import logging
from typing import Any, Awaitable, Callable, List, Type

from .abc import ManagerAPI, ServiceAPI, _InternalManagerAPI
from .exceptions import LifecycleError
from .typing import EXC_INFO


class Service(ServiceAPI):
    pass


LogicFnType = Callable[..., Awaitable[Any]]


def as_service(service_fn: LogicFnType) -> Type[ServiceAPI]:
    """
    Create a service out of a simple function
    """

    class _Service(Service):
        def __init__(self, *args: Any, **kwargs: Any):
            self._args = args
            self._kwargs = kwargs

        async def run(self) -> None:
            await service_fn(self.manager, *self._args, **self._kwargs)

    _Service.__name__ = service_fn.__name__
    _Service.__doc__ = service_fn.__doc__
    return _Service


class BaseManager(_InternalManagerAPI):
    logger = logging.getLogger("p2p.service.Manager")

    _service: ServiceAPI

    _errors: List[EXC_INFO]

    def __init__(self, service: ServiceAPI) -> None:
        if hasattr(service, "manager"):
            raise LifecycleError("Service already has a manager.")
        else:
            service.manager = self

        self._service = service

        # errors
        self._errors = []

    #
    # Event API mirror
    #
    @property
    def is_running(self) -> bool:
        return self.is_started and not (self.is_stopping or self.is_finished)

    @property
    def did_error(self) -> bool:
        return len(self._errors) > 0

    #
    # Control API
    #
    async def stop(self) -> None:
        self.cancel()
        await self.wait_finished()

    #
    # Wait API
    #
    def run_daemon_task(
        self, async_fn: Callable[..., Awaitable[Any]], *args: Any, name: str = None
    ) -> None:

        self.run_task(async_fn, *args, daemon=True, name=name)

    def run_daemon_child_service(
        self, service: ServiceAPI, name: str = None
    ) -> ManagerAPI:
        return self.run_child_service(service, daemon=True, name=name)



import asyncio
import functools
import sys
from typing import (
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    Coroutine,
    Dict,
    List,
    Set,
    TypeVar,
    cast,
)

from async_generator import asynccontextmanager
from trio import MultiError

from async_service._utils import iter_dag

from .abc import ManagerAPI, ServiceAPI
from .base import BaseManager
from .exceptions import DaemonTaskExit, LifecycleError, ServiceCancelled
from .typing import EXC_INFO


class AsyncioManager(BaseManager):
    # Tracking of the system level background tasks.
    _system_tasks: Set["asyncio.Future[None]"]

    # Tracking of the background tasks that the service has initiated.
    _service_task_dag: Dict["asyncio.Future[Any]", List["asyncio.Future[Any]"]]

    def __init__(
        self, service: ServiceAPI, loop: asyncio.AbstractEventLoop = None
    ) -> None:
        super().__init__(service)

        self._loop = loop

        # events
        self._started = asyncio.Event()
        self._cancelled = asyncio.Event()
        self._stopping = asyncio.Event()
        self._finished = asyncio.Event()

        # locks
        self._run_lock = asyncio.Lock()

        # task tracking
        self._service_task_dag = {}
        self._system_tasks = set()

    #
    # System Tasks
    #
    async def _handle_cancelled(self) -> None:
        """
        When cancellation is requested this triggers the cancellation of all
        background tasks.
        """
        self.logger.debug("%s: _handle_cancelled waiting for cancellation", self)
        await self.wait_cancelled()
        self.logger.debug("%s: _handle_cancelled triggering task cancellation", self)

        # TODO: need new comment here explaining the way we iterate in
        # dependency first order.
        for task in iter_dag(self._service_task_dag):
            if not task.done():
                task.cancel()

            try:
                await task
            except asyncio.CancelledError:
                pass
            except Exception:
                self._errors.append(cast(EXC_INFO, sys.exc_info()))

    async def _handle_stopping(self) -> None:
        """
        Once the `_stopping` event is set this triggers cancellation of the system tasks.
        """
        self.logger.debug("%s: _handle_stopping waiting for stopping", self)
        await self.wait_stopping()
        self.logger.debug(
            "%s: _handle_stopping triggering system task cancellations", self
        )

        # trigger cancellation of all of the system tasks
        for task in self._system_tasks:
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass
            except Exception:
                self._errors.append(cast(EXC_INFO, sys.exc_info()))

    @classmethod
    async def run_service(
        cls, service: ServiceAPI, loop: asyncio.AbstractEventLoop = None
    ) -> None:
        manager = cls(service, loop=loop)
        await manager.run()

    async def run(self) -> None:
        if self._run_lock.locked():
            raise LifecycleError(
                "Cannot run a service with the run lock already engaged.  Already started?"
            )
        elif self.is_started:
            raise LifecycleError("Cannot run a service which is already started.")

        async with self._run_lock:
            try:
                handle_cancelled_task = asyncio.ensure_future(
                    self._handle_cancelled(), loop=self._loop
                )
                handle_stopping_task = asyncio.ensure_future(
                    self._handle_stopping(), loop=self._loop
                )

                self._system_tasks.add(handle_cancelled_task)
                self._system_tasks.add(handle_stopping_task)

                self.run_task(self._service.run, _root=True)

                self._started.set()

                # block here until all service tasks have finished
                await self._wait_service_tasks()
            finally:
                # signal that the service is stopping
                self._stopping.set()
                self.logger.debug("%s stopping", self)

            # block here until all system tasks have finished.
            await asyncio.wait(
                (handle_cancelled_task, handle_stopping_task),
                return_when=asyncio.ALL_COMPLETED,
            )

        self._finished.set()
        self.logger.debug("%s finished", self)

        # If an error occured, re-raise it here
        if self.did_error:
            raise MultiError(
                tuple(
                    exc_value.with_traceback(exc_tb)
                    for _, exc_value, exc_tb in self._errors
                )
            )

    async def _wait_service_tasks(self) -> None:
        while True:
            done, pending = await asyncio.wait(
                tuple(self._service_task_dag.keys()), return_when=asyncio.ALL_COMPLETED
            )
            if all(task.done() for task in self._service_task_dag.keys()):
                break

    #
    # Event API mirror
    #
    @property
    def is_started(self) -> bool:
        return self._started.is_set()

    @property
    def is_cancelled(self) -> bool:
        return self._cancelled.is_set()

    @property
    def is_stopping(self) -> bool:
        return self._stopping.is_set() and not self.is_finished

    @property
    def is_finished(self) -> bool:
        return self._finished.is_set()

    #
    # Control API
    #
    def cancel(self) -> None:
        if not self.is_started:
            raise LifecycleError("Cannot cancel as service which was never started.")
        self._cancelled.set()

    #
    # Wait API
    #
    async def wait_started(self) -> None:
        await self._started.wait()

    async def wait_cancelled(self) -> None:
        await self._cancelled.wait()

    async def wait_stopping(self) -> None:
        await self._stopping.wait()

    async def wait_finished(self) -> None:
        await self._finished.wait()

    async def _run_and_manage_task(
        self,
        async_fn: Callable[..., Awaitable[Any]],
        *args: Any,
        daemon: bool,
        name: str,
    ) -> None:
        try:
            await async_fn(*args)
        except asyncio.CancelledError:
            raise
        except Exception as err:
            self.logger.debug(
                "task '%s[daemon=%s]' exited with error: %s",
                name,
                daemon,
                err,
                exc_info=True,
            )
            self._errors.append(cast(EXC_INFO, sys.exc_info()))
            self.cancel()
        else:
            self.logger.debug("task '%s[daemon=%s]' finished.", name, daemon)
            if daemon:
                self.logger.debug(
                    "daemon task '%s' exited unexpectedly.  Cancelling service: %s",
                    name,
                    self,
                )
                self.cancel()
                raise DaemonTaskExit(f"Daemon task {name} exited")

    def run_task(
        self,
        async_fn: Callable[..., Awaitable[Any]],
        *args: Any,
        daemon: bool = False,
        name: str = None,
        _root: bool = False,
    ) -> None:
        current_task = asyncio.Task.current_task()
        if not _root and current_task not in self._service_task_dag:
            raise Exception(f"TODO: unknown task {current_task}")
        task = asyncio.ensure_future(
            self._run_and_manage_task(
                async_fn, *args, daemon=daemon, name=name or repr(async_fn)
            ),
            loop=self._loop,
        )
        self._service_task_dag[task] = []
        if not _root:
            self._service_task_dag[current_task].append(task)

    def run_child_service(
        self, service: ServiceAPI, daemon: bool = False, name: str = None
    ) -> ManagerAPI:
        child_manager = type(self)(service, loop=self._loop)
        self.run_task(child_manager.run, daemon=daemon, name=name or repr(service))
        return child_manager


TReturn = TypeVar("TReturn")
TFunc = TypeVar("TFunc", bound=Callable[..., Coroutine[Any, Any, TReturn]])


def external_api(func: TFunc) -> TFunc:
    @functools.wraps(func)
    async def inner(self: ServiceAPI, *args: Any) -> TReturn:
        if self.manager.is_stopping:
            raise ServiceCancelled(
                f"Cannot access external API {func}.  Service is already cancelled"
            )

        func_task: "asyncio.Future[TReturn]" = asyncio.ensure_future(func(self, *args))
        service_stopping_task = asyncio.ensure_future(self.manager.wait_stopping())

        done, pending = await asyncio.wait(
            (func_task, service_stopping_task), return_when=asyncio.FIRST_COMPLETED
        )

        if func_task.done():
            service_stopping_task.cancel()
            try:
                await service_stopping_task
            except asyncio.CancelledError:
                pass
            return await func_task
        elif service_stopping_task.done():
            func_task.cancel()
            try:
                await func_task
            except asyncio.CancelledError:
                pass
            raise ServiceCancelled(
                f"Cannot access external API {func}.  Service is already cancelled"
            )
        else:
            raise Exception("Code path should be unreachable")

    return cast(TFunc, inner)


@asynccontextmanager
async def background_asyncio_service(
    service: ServiceAPI, loop: asyncio.AbstractEventLoop = None
) -> AsyncIterator[ManagerAPI]:
    """
    Run a service in the background.

    The service is running within the context block and will be properly
    cleaned up upon exiting the context block.
    """
    manager = AsyncioManager(service, loop=loop)
    task = asyncio.ensure_future(manager.run(), loop=loop)

    try:
        await manager.wait_started()

        try:
            yield manager
        finally:
            await manager.stop()

        assert not manager.did_error, "ARST ARST ARST"

    finally:
        task.cancel()

        try:
            await task
        except asyncio.CancelledError:
            pass

        if manager.did_error:
            # TODO: better place for this.
            raise MultiError(
                tuple(
                    exc_value.with_traceback(exc_tb)
                    for _, exc_value, exc_tb in manager._errors
                )
            )
